/// SinkWorld split buffer template class
/// Copyright 2001 Neil Hodgson
/// All implementations are expanded over the <T> type and <N> number of bytes used for names.

#include <string.h>

#include "SW.h"
#include "SWDefs.h"
#include "IStyles.h"
#include "cbx_<N>.h"

/// Move the gap to a particular position so that insertion and 
/// deletion at that point will not require much copying and 
/// hence be fast.
/// TODO: for vectors of pointers, it may be worth while filling the gap with nulls.
void cbx_<N>::GapTo(int position) {
	if (position != part1Length) {
		if (position < part1Length) {
			SW_ARRAYCOPY(body, position, 
				body, position + gapLength,
				part1Length - position);
		} else {	// position > part1Length
			SW_ARRAYCOPY(body, part1Length + gapLength, 
				body, part1Length,
				position - part1Length);
		}
		part1Length = position;
	}
}

/// Reallocate the storage for the buffer to be newSize and 
/// copy exisiting contents to the new buffer.
/// Must not be used to decrease the size of the buffer.
void cbx_<N>::ReAllocate(int newSize) {
	// Move the gap to the end
	GapTo(lengthBody);
	<T> *newBody = new <T>[newSize];
	if ((size != 0) && (body != NULL)) {
		SW_ARRAYCOPY(body, 0, newBody, 0, lengthBody);
		delete []body;
	}
	body = newBody;
	gapLength += newSize - size;
	size = newSize;
}

/// Check that there is room in the buffer for an insertion, 
/// reallocating if more space needed.
void cbx_<N>::RoomFor(int insertionLength) {
	if (gapLength <= insertionLength) {
		if (growSize * 6 < size)
			growSize *= 2;
		ReAllocate(size + insertionLength + growSize);
	}
}

/// Construct a split buffer.
cbx_<N>::cbx_<N>(int initialLength, int growSize_) {
	growSize = growSize_;
	size = 0;
	lengthBody = 0;
	part1Length = 0;
	gapLength = 0;
	ReAllocate(initialLength);
}

cbx_<N>::~cbx_<N>() {
	delete []body;
	body = NULL;
}

int cbx_<N>::GetGrowSize() {
	return growSize;
}

void cbx_<N>::SetGrowSize(int growSize_) {
	growSize = growSize_;
}

int cbx_<N>::ValueAt(int position) {
	int ch = CharAt(position);
	// For bytes want to return as if unsigned so can be used as an index.
	if ((<N> == 1) && (ch < 0))
		ch += 1 << (8 * <N>);
	return ch;
}

void cbx_<N>::SetValueAt(int position, int value) {
	if (position < part1Length) {
		if (position < 0) {
			throw new BadPositionException();
		} else {
			body[position] = (<T>)value;
		}
	} else {
		if (position >= lengthBody) {
			throw new BadPositionException();
		} else {
			body[gapLength + position] = (<T>)value;
		}
	}
}

/// Retrieve the length of the buffer.
int cbx_<N>::Length() {
	return lengthBody;
}

/// Retrieve a range of text from the buffer.
/// Retrieving positions outside the range of the buffer throws a
/// BadPositionException.
void cbx_<N>::RetrieveString(int position, <T> s[], int retrieveLength) {
	if ((position < 0) || ((position+retrieveLength) > lengthBody)) {
		throw new BadPositionException();
	}
	int i = 0;
	while ((i < retrieveLength) && (position < part1Length)) {
		s[i++] = body[position++];
	}
	while (i < retrieveLength) {
		s[i++] = body[gapLength + position++];
	}
}

/// Insert text into the buffer.
/// Inserting at positions outside the current range throws a
/// BadPositionException.
void cbx_<N>::InsertString(int position, <T> s[], int insertLength) {
	InsertFromArray(position, s, 0, insertLength);
}

/// Insert text into the buffer from an array.
void cbx_<N>::InsertFromArray(int positionToInsert, <T> s[], int positionFrom, int insertLength) {
	if (insertLength > 0) {
		if ((positionToInsert < 0) || (positionToInsert > lengthBody)) {
			throw new BadPositionException();
		}
		RoomFor(insertLength);
		GapTo(positionToInsert);
		SW_ARRAYCOPY(s, positionFrom, body, part1Length, insertLength);
		lengthBody += insertLength;
		part1Length += insertLength;
		gapLength -= insertLength;
	}
}

/// Append text to the buffer.
void cbx_<N>::AppendString(<T> s[], int insertLength) {
	InsertString(lengthBody, s, insertLength);
}

/// Append text to the buffer from another buffer.
void cbx_<N>::AppendFromBuffer(cbx_<N> *other, int position, int insertLength) {
	
	if ((position < 0) || ((position + insertLength) > other->lengthBody)) {
		throw new BadPositionException();
	}
	// Break into two ranges around the other buffers gap
	int lengthBeforeGap = insertLength;
	if (position > other->part1Length) {
		lengthBeforeGap = 0;
	} else if ((position + insertLength) > other->part1Length) {
		lengthBeforeGap = other->part1Length - position;
	}
	InsertFromArray(lengthBody, other->body, position, lengthBeforeGap);
	int pos = position + lengthBeforeGap + other->gapLength;
	InsertFromArray(lengthBody, other->body, pos, insertLength - lengthBeforeGap);
}

/// Fill an already allocated range with a value.
/// Modifying outside the current range throws a
/// BadPositionException.
void cbx_<N>::FillRange(int position, int val, int fillLength) {
	if (fillLength > 0) {
		if ((position < 0) || ((position + fillLength) > lengthBody)) {
			throw new BadPositionException();
		}
		// Several optimizations tried here, including block filling functions
		// and simplifying the loops to only have one changing variable but no 
		// change made any real impact.
		int i = 0;
		int pos = position;
		while ((i < fillLength) && (pos < part1Length)) {
			body[pos++] = (<T>)val;
			i++;
		}
		while (i < fillLength) {
			body[gapLength + pos++] = (<T>)val;
			i++;
		}
	}
}

/// Insert a number of zero characters into the buffer.
/// Inserting at positions outside the current range fails.
void cbx_<N>::InsertZeros(int position, int insertLength) {
	if (insertLength > 0) {
		RoomFor(insertLength);
		GapTo(position);
		for (int i = 0; i < insertLength; i++)
			body[part1Length + i] = 0;
		lengthBody += insertLength;
		part1Length += insertLength;
		gapLength -= insertLength;
	}
}

/// Delete a range from the buffer.
/// Deleting positions outside the current range fails.
void cbx_<N>::DeleteRange(int position, int deleteLength) {
	if (deleteLength > 0) {
		GapTo(position);
		lengthBody -= deleteLength;
		gapLength += deleteLength;
	}
}
