#include <string.h>

#include "SW.h"
#include "SWDefs.h"
#include "IStyles.h"
#include "cbx_0.h"
#include "cbx_1.h"
#include "cbx_2.h"
#include "cbx_4.h"
#include "lv.h"
#include "undo.h"
#include "LineData.h"
#include "pbx.h"

void pbx::InsertLine(int lineInsert, int position) {
	lines->InsertLine(lineInsert, position);
	LineData *pld = pldRoot;
	while (pld != NULL) {
		pld->InsertLine(lineInsert);
		pld = pld->next;
	}
}

void pbx::RemoveLine(int lineRemove) {
	lines->RemoveLine(lineRemove);
	LineData *pld = pldRoot;
	while (pld != NULL) {
		pld->RemoveLine(lineRemove);
		pld = pld->next;
	}
}

void pbx::BasicInsertString(int position, SW_BYTE s[], int insertLength) {
	substance->InsertString(position, s, insertLength);
	style->InsertZeros(position, insertLength);
	
	if (lines != NULL) {
		int lineInsert = lines->LineFromPosition(position) + 1;
		lines->InsertText(lineInsert, insertLength);
		SW_BYTE chPrev = (SW_BYTE)' ';
		if ((position - 1) >= 0)
			chPrev = CharAt(position - 1);
		SW_BYTE chAfter = CharAt(position + insertLength);
		if (chPrev == '\r' && chAfter == '\n') {
			// Splitting up a crlf pair at position
			InsertLine(lineInsert, position);
			lineInsert++;
		}
		SW_BYTE ch = (SW_BYTE)' ';
		for (int i = 0; i < insertLength; i ++) {
			ch = s[i];
			if (ch == '\r') {
				if ((i != (insertLength-1)) || (chAfter != '\n')) {
					InsertLine(lineInsert, (position + i) + 1);
					lineInsert++;
				}
			} else if (ch == '\n') {
				if (chPrev == '\r') {
					// Patch up what was end of line
					lines->SetLineStartPosition(lineInsert - 1, (position + i) + 1);
				} else {
					InsertLine(lineInsert, (position + i) + 1);
					lineInsert++;
				}
			}
			chPrev = ch;
		}
	}
}

void pbx::BasicDeleteChars(int position, int deleteLength) {
	// Have to fix up line positions before doing deletion as looking at text in buffer
	// to work out which lines have been removed

	if (lines != NULL) {
		int lineRemove = lines->LineFromPosition(position) + 1;
		// Point all the lines after the insertion point less further along in the buffer
		lines->InsertText(lineRemove, -deleteLength);
		SW_BYTE chPrev = (SW_BYTE)' ';
		if (position >= 1)
			chPrev = CharAt(position - 1);
		SW_BYTE chBefore = chPrev;
		SW_BYTE chNext = (SW_BYTE)' ';
		if (position < Length())
			chNext = CharAt(position);
		int i = 0;
		if (chPrev == '\r' && chNext == '\n') {
			// Breaking apart a \r\n pair
			// Move back one
			lines->SetLineStartPosition(lineRemove, position);
			lineRemove++;
			// First \n is not real deletion so skip it
			chNext = CharAt(position + i + 1);
			i++;
		}

		SW_BYTE ch = chNext;
		for (;i < deleteLength; i++) {
			chNext = CharAt(position + i + 1);
			if (ch == '\r') {
				if (chNext != '\n') {
					RemoveLine(lineRemove);
				}
			} else if (ch == '\n') {
				RemoveLine(lineRemove);
			}
			ch = chNext;
		}

		// May have to fix up end if last deletion causes cr to be next to lf
		// or removes one of a crlf pair
		SW_BYTE chAfter = (SW_BYTE)' ';
		if ((position + deleteLength) < Length())
			chAfter = CharAt(position + deleteLength);
		if (chBefore == '\r' && chAfter == '\n') {
			// Using lineRemove-1 as cr ended line before start of deletion
			RemoveLine(lineRemove - 1);
			lines->SetLineStartPosition(lineRemove - 1, position + 1);
		}
	}
	
	substance->DeleteRange(position, deleteLength);
	style->DeleteRange(position, deleteLength);
}

pbx::pbx(int initialLength, int growSize_, int styleWidth_) {
	styleWidth = styleWidth_;
	substance = new cbx_1(initialLength, growSize_);
	if (styleWidth == 0)
		style = new cbx_0();
	else if (styleWidth == 1)
		style = new cbx_1(initialLength, growSize_);
	else if (styleWidth == 2)
		style = new cbx_2(initialLength, growSize_);
	else if (styleWidth == 4)
		style = new cbx_4(initialLength, growSize_);
	lines = new lv(growSize_);
	us = new undoStack(1000, 1000);
	pldRoot = NULL;
	encoding = ENCODING_BYTE;
	leads = NULL;
}

pbx::~pbx() {
	delete substance;
	substance = NULL;
	delete style;
	style = NULL;
	delete lines;
	lines = NULL;
	delete us;
	us = NULL;
	delete []leads;
	leads = NULL;
}

int pbx::StyleAt(int position) {
	return style->ValueAt(position);
}

int pbx::Length() {
	return substance->Length();
}

/// Retrieving positions outside the range of the buffer fails
void pbx::RetrieveString(int position, SW_BYTE s[], int retrieveLength) {
	substance->RetrieveString(position, s, retrieveLength);
}

void pbx::DiscardLines() {
	delete lines;
	lines = NULL;
}

void pbx::RecreateLines() {
	delete lines;
	lines = new lv(128);
	SW_BYTE chPrev = (SW_BYTE)' ';
	int len = substance->Length();
	int lineInsert = 1;
	lines->InsertText(lineInsert, len);
	for (int i = 0; i < len; i ++) {
		SW_BYTE ch = substance->CharAt(i);
		if (ch == '\r') {
			InsertLine(lineInsert, i + 1);
			lineInsert++;
		} else if (ch == '\n') {
			if (chPrev == '\r') {
				// Patch up what was end of line
				lines->SetLineStartPosition(lineInsert - 1, i + 1);
			} else {
				InsertLine(lineInsert, i + 1);
				lineInsert++;
			}
		}
		chPrev = ch;
	}
}

void pbx::InsertString(int position, SW_BYTE s[], int insertLength) {
	if (us != NULL) {
		us->AppendAction(undoStack::ACTION_INSERT, position, NULL, insertLength);
	}

	BasicInsertString(position, s, insertLength);
}

void pbx::FillStyleRange(int position, int val, int fillLength) {
	style->FillRange(position, val, fillLength);
}	

void pbx::DeleteRange(int position, int deleteLength) {
	if (us != NULL) {
		us->AppendAction(undoStack::ACTION_REMOVE, position, substance, deleteLength);
	}
	BasicDeleteChars(position, deleteLength);
}

int pbx::Lines() {
	if (lines == NULL)
		RecreateLines();
	return lines->Lines();
}

int pbx::PositionFromLine(int line) {
	if (lines == NULL)
		RecreateLines();
	return lines->PositionFromLine(line);
}

int pbx::LineEndPositionFromLine(int line) {
	if (line == Lines()-1)
		return Length();
	else 
		return PositionFromLine(line+1);
}

int pbx::LineFromPosition(int position) {
	if (lines == NULL)
		RecreateLines();
	return lines->LineFromPosition(position);
}

void pbx::AddPerLineData(LineData *pld) {
	pld->next = pldRoot;
	pldRoot = pld;
	// Synchronize to current number of lines
	for (int i=0; i < Lines(); i++) {
		pld->InsertLine(i);
	}
}

void pbx::RemovePerLineData(LineData *pld) {
	if (pldRoot != NULL) {
		if (pldRoot == pld) {
			pldRoot = pld->next;
			pld->next = NULL;
		} else {
			LineData *pldBase = pldRoot;
			while ((pldBase->next != NULL) && (pldBase->next != pld)) {
				pldBase = pldBase->next;
			}
			if (pldBase->next != NULL) {
				pldBase->next = pld->next;
				pld->next = NULL;
			}
		}
	}
}

void pbx::SetUndoCollection(bool collectUndo) {
	if (collectUndo) {
		if (us == NULL)
			us = new undoStack(1000, 1000);
		us->DropUndoSequence();
	} else {
		delete us;
		us = NULL;
	}
}

bool pbx::IsCollectingUndo() {
	return us != NULL;
}

void pbx::BeginUndoAction() {
	us->BeginUndoAction();
}

void pbx::EndUndoAction() {
	us->EndUndoAction();
}

void pbx::DeleteUndoHistory() {
	delete us;
	us = new undoStack(1000, 1000);
}

bool pbx::CanUndo() {
	return us->CanUndo();
}

int pbx::StartUndo() {
	return us->StartUndo();
}

undoAction *pbx::GetUndoStep() {
	undoAction *action = us->GetUndoStep();
	int at = action->actionType;
	at &= ~undoStack::ACTION_START;
	action->data = new SW_BYTE[action->lenData];
	if (undoStack::ACTION_INSERT == at) {
		substance->RetrieveString(action->position, action->data, action->lenData);
		// Add to inserts so can be redone
		us->AppendInsertDataStep(action);
	} else {
		us->RetrieveRemoveDataStep(action);
	}
	return action;
}

void pbx::PerformUndoAction() {
	undoAction *action = GetUndoStep();
	if (action->actionType == undoStack::ACTION_INSERT) {
		BasicDeleteChars(action->position, action->lenData);
	} else if (action->actionType == undoStack::ACTION_REMOVE) {
		BasicInsertString(action->position, action->data, action->lenData);
	}
	us->CompletedUndoStep();
	delete action;
}

void pbx::PerformUndoStep() {
	int steps = StartUndo();
	for (int i=0;i<steps;i++)
		PerformUndoAction();
}

bool pbx::CanRedo() {
	return us->CanRedo();
}

int pbx::StartRedo() {
	return us->StartRedo();
}

undoAction *pbx::GetRedoStep() {
	return us->GetRedoStep();
}

void pbx::PerformRedoAction() {
	undoAction *action = GetRedoStep();
	if (action->actionType == undoStack::ACTION_INSERT) {
		BasicInsertString(action->position, action->data, action->lenData);
	} else if (action->actionType == undoStack::ACTION_REMOVE) {
		BasicDeleteChars(action->position, action->lenData);
	}
	us->CompletedRedoStep();
	delete action;
}

void pbx::PerformRedoStep() {
	int steps = StartRedo();
	for (int i=0;i<steps;i++)
		PerformRedoAction();
}

int pbx::ClampPositionIntoDocument(int position) {
	int len = Length();
	if (position > len)
		return len;
	if (position < 0)
		return 0;
	return position;
}

bool pbx::IsCrLf(int position) {
	if (position < 0)
		return false;
	if (position >= (Length() - 1))
		return false;
	return (CharAt(position) == '\r') && (CharAt(position + 1) == '\n');
}

bool pbx::IsDBCS(int position) {
	// Anchor DBCS calculations at start of line because start of line can
	// not be a DBCS trail byte.
	int startLine = position;
	while (startLine > 0 && CharAt(startLine) != '\r' && CharAt(startLine) != '\n') {
		startLine--;
	}
	while (startLine <= position) {
		int ch = CharAt(startLine);
		if (ch < 0)
			ch += 256;
		if (leads[ch]) {
			startLine++;
			if (startLine >= position)
				return true;
		}
	}
	return false;
}

int pbx::LenChar(int position) {
	if (IsCrLf(position)) {
		return 2;
	} else if (encoding == ENCODING_UTF8) {
		int u = CharAt(position);
		if (u < 0)
			u += 256;
		if (u < 0x80) {
			return 1;
		} else if (u < (0x80 + 0x40 + 0x20)) {
			return 2;
		} else {
			return 3;
		}
	} else if ((encoding == ENCODING_DBCS) && IsDBCS(position)) {
		return 2;
	} else {
		return 1;
	}
}

int pbx::CharValue(int position) {
	// Convert to unsigned
	int u1 = CharAt(position);
	if (u1 < 0)
		u1 += 256;
	switch (encoding) {
		case ENCODING_BYTE:
			return u1;
		case ENCODING_DBCS:
			if (leads[u1]) {
				int u2 = CharAt(position+1) ;
				if (u2 < 0)
					u2 += 256;
				return (u1 << 8) + u2;
			} else {
				return u1;
			}
		case ENCODING_UTF8:
		default:
			return ((u1 & 0xF) << 12) + 
				((CharAt(position+1) & 0x7F) << 6) + 
				(CharAt(position+2) & 0x7F);
	}
}

// Normalise a position so that it is not halfway through a two byte character.
// This can occur in two situations -
// When lines are terminated with \r\n pairs which should be treated as one character.
// When displaying DBCS text such as Japanese.
// If moving, move the position in the indicated direction.
int pbx::MovePositionOutsideChar(int position, int moveDirection, bool checkLineEnd) {
	// If out of range, just return value - should be fixed up after
	if (position < 0)
		return position;
	if (position > Length())
		return position;

	// Position 0 and Length() can not be between any two characters
	if (position == 0)
		return position;
	if (position == Length())
		return position;

	// assert position > 0 && position < Length()
	if (checkLineEnd && IsCrLf(position - 1)) {
		if (moveDirection > 0)
			return position + 1;
		else
			return position - 1;
	}

	// Not between CR and LF
	if (encoding == ENCODING_UTF8) {
		int ch = CharAt(position);
		if (ch < 0)
			ch += 256;
		while ((position > 0) && (position < Length()) && (ch >= 0x80) && (ch < (0x80 + 0x40))) {
			// ch is a trail byte
			if (moveDirection > 0)
				position++;
			else
				position--;
			ch = CharAt(position);
			if (ch < 0)
				ch += 256;
		}
	} else if (encoding == ENCODING_DBCS) {
		// Anchor DBCS calculations at start of line because start of line can
		// not be a DBCS trail byte.
		int startLine = position;
		while (startLine > 0 && CharAt(startLine) != '\r' && CharAt(startLine) != '\n') {
			startLine--;
		}
		bool atLeadByte = false;
		while (startLine < position) {
			if (atLeadByte) {
				atLeadByte = false;
			} else {
				int ch = CharAt(startLine);
				if (ch < 0)
					ch += 256;
				atLeadByte = leads[ch];
			}
			startLine++;
		}


		if (atLeadByte) {
			// Position is between a lead byte and a trail byte
			if (moveDirection > 0)
				return position + 1;
			else
				return position - 1;
		}
	}

	return position;
}
